import datetime as dt
import io
import json
import pickle
from typing import Any

import pulsar
from fastavro import parse_schema, schemaless_writer, schemaless_reader

from kikyo import Kikyo
from kikyo.datahub import DataHub, Producer, Consumer, Message

record_schema = {
    'name': 'DataHubRecord',
    'namespace': 'kikyo.datahub',
    'type': 'record',
    'fields': [
        {'name': 'type', 'type': 'string'},
        {'name': 'data', 'type': 'bytes'}
    ]
}

parsed_record_schema = parse_schema(record_schema)


class PulsarBasedDataHub(DataHub):
    def __init__(self, client: Kikyo):
        settings = client.settings.deep('pulsar')
        if not settings:
            return

        self.tenant = settings.get('tenant', 'public')
        self.namespace = settings.get('namespace', 'default')
        self.pulsar = pulsar.Client(settings['service_url'])

        client.add_component('pulsar_datahub', self)

    def create_producer(self, topic: str) -> Producer:
        return PulsarBasedProducer(self, topic)

    def subscribe(
            self,
            topic: str,
            subscription_name: str = None,
            consumer_name: str = None,
            auto_ack: bool = True,
    ) -> Consumer:
        return PulsarBasedConsumer(
            self,
            topic,
            subscription_name=subscription_name,
            consumer_name=consumer_name,
            auto_ack=auto_ack,
        )

    def get_topic(self, name: str):
        return f'persistent://{self.tenant}/{self.namespace}/{name}'


class PulsarBasedProducer(Producer):
    def __init__(self, datahub: PulsarBasedDataHub, topic: str):
        super().__init__()
        self.producer = datahub.pulsar.create_producer(
            datahub.get_topic(topic),
            block_if_queue_full=True,
        )

    def send(
            self,
            record: Any,
            partition_key: str = None,
    ):
        data = MessageUtils.build(record)
        self.producer.send(
            data,
            partition_key=partition_key,
        )

    def close(self):
        self.producer.close()


class PulsarMessage(Message):
    def __init__(self, msg: pulsar.Message):
        self._msg = msg
        self._value = MessageUtils.extract_data(msg.data())
        self._publish_time = None

    @property
    def value(self) -> Any:
        return self._value

    @property
    def publish_time(self) -> dt.datetime:
        if self._publish_time is None:
            self._publish_time = dt.datetime.fromtimestamp(self._msg.publish_timestamp() / 1000)
        return self._publish_time


class PulsarBasedConsumer(Consumer):
    def __init__(
            self,
            datahub: PulsarBasedDataHub,
            topic: str,
            subscription_name: str = None,
            consumer_name: str = None,
            auto_ack: bool = True,
    ):
        super().__init__()
        self.consumer = datahub.pulsar.subscribe(
            datahub.get_topic(topic),
            consumer_type=pulsar.ConsumerType.KeyShared,
            subscription_name=subscription_name,
            consumer_name=consumer_name,
            initial_position=pulsar.InitialPosition.Earliest,
        )
        self._auto_ack = auto_ack

    def receive(self) -> Message:
        msg = self.consumer.receive()
        if self._auto_ack:
            self.consumer.acknowledge(msg)
        return PulsarMessage(msg)

    def close(self):
        self.consumer.close()

    def ack(self, msg: Message):
        assert isinstance(msg, PulsarMessage)
        self.consumer.acknowledge(msg._msg)


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, dt.datetime):
            return obj.isoformat()
        elif isinstance(obj, dt.date):
            return obj.isoformat()
        return super().default(obj)


class MessageUtils:
    ENCODING = 'utf-8'

    @classmethod
    def build(cls, data: Any) -> bytes:
        if isinstance(data, (dict, list)):
            dtype = 'json'
            data = json.dumps(
                data,
                ensure_ascii=False,
                cls=JSONEncoder,
            ).encode(encoding=cls.ENCODING)
        elif isinstance(data, bytes):
            dtype = 'bytes'
            data = data
        elif isinstance(data, str):
            dtype = 'str'
            data = data.encode(encoding=cls.ENCODING)
        else:
            dtype = 'object'
            data = pickle.dumps(data)

        d = {
            'type': dtype,
            'data': data,
        }
        wio = io.BytesIO()
        schemaless_writer(wio, parsed_record_schema, d)
        return wio.getvalue()

    @classmethod
    def extract_data(cls, content: bytes) -> Any:
        message = schemaless_reader(io.BytesIO(content), parsed_record_schema)
        if message['type'] == 'json':
            return json.loads(message['data'].decode(encoding=cls.ENCODING))
        if message['type'] == 'bytes':
            return message['data']
        if message['type'] == 'str':
            return message['data'].decode(encoding=cls.ENCODING)
        if message['type'] == 'object':
            return pickle.loads(message['data'])



from typing import List, Optional, Any, Sequence, Dict

from elasticsearch import Elasticsearch, helpers
from elasticsearch import RequestsHttpConnection
from pydantic import BaseModel

from kikyo import Kikyo
from kikyo.search import Search, Index


class BaseResponse:
    def __init__(self, resp: dict):
        self.response = resp

    def __getitem__(self, key):
        return self.response.get(key)


class SearchResponse(BaseResponse):
    def hits_total(self) -> Optional[int]:
        total = self['hits'].get('total')
        if total:
            return total['value']

    def hits(self) -> List[dict]:
        return [h['_source'] for h in self['hits']['hits']]

    def raw_hits(self) -> List[dict]:
        return self['hits']['hits']

    def count(self) -> int:
        return self['count']

    def buckets(self, name: str) -> List[dict]:
        return self['aggregations'][name]['buckets']


class BulkResponse(BaseResponse):
    def errors(self) -> dict:
        return self['errors']

    def success(self) -> int:
        return self['success']


class EsBasedSearch(Search):
    def __init__(self, client: Kikyo):
        settings = client.settings.deep('elasticsearch')

        if not settings:
            return

        kwargs = {
            'connection_class': RequestsHttpConnection,
            'hosts': settings['hosts']
        }
        if 'username' in settings:
            kwargs['http_auth'] = (settings['username'], settings['password'])
        if 'timeout' in settings:
            kwargs['timeout'] = settings['timeout']
        if 'extra_params' in settings:
            kwargs.update(settings['extra_params'])

        self.es = Elasticsearch(**kwargs)
        self.index_prefix = settings.get('index_prefix', default='')

        client.add_component('es_search', self)

    def get_index_of_topic(self, topic):
        if ',' in topic:
            topics = topic.split(',')
            topic = ','.join([t if t.startswith(self.index_prefix) else f'{self.index_prefix}{t}' for t in topics])
            return topic
        if topic.startswith(self.index_prefix):
            return topic
        return f'{self.index_prefix}{topic}'

    def index(self, topic: str) -> 'EsIndex':
        return EsIndex(topic, self)

    def search(self, index: str) -> 'IndexedQuery':
        """
        指定索引

        :param index: index名称
        """
        return IndexedQuery(index, self)


class EsIndex(Index):
    def __init__(self, topic: str, client: EsBasedSearch):
        self.topic = topic
        self.es = client.es
        self.client = client

    def exists(self, id: str) -> bool:
        return self.es.exists(
            self.client.get_index_of_topic(self.topic),
            id,
        )

    def get(self, id: str) -> dict:
        resp = self.es.get(
            self.client.get_index_of_topic(self.topic),
            id,
        )
        return resp['_source']

    def mget(self, *ids: List[str]) -> List[dict]:
        resp = self.es.mget(
            index=self.client.get_index_of_topic(self.topic),
            body={
                'ids': ids,
            }
        )
        return [(i['_source'] if i['found'] else None) for i in resp['docs']]

    def put(self, id: str, data: dict, refresh: bool = None):
        """
        更新指定数据，指定ID不存在时自动创建数据

        :param id: 数据ID
        :param data: 数据内容
        :param refresh: 强制刷新索引
        """
        kwargs = {}
        if refresh is not None:
            kwargs['refresh'] = refresh

        self.es.index(
            self.client.get_index_of_topic(self.topic),
            body=data,
            id=id,
            **kwargs,
        )

    def delete(self, id: str, refresh: bool = None):
        """
        删除指定数据

        :param id: 数据ID
        :param refresh: 强制刷新索引
        """
        kwargs = {}
        if refresh is not None:
            kwargs['refresh'] = refresh

        self.es.delete(
            self.client.get_index_of_topic(self.topic),
            id,
            **kwargs,
        )

    def update(self, id: str, data: dict = None, refresh: bool = None):
        """
        更新数据的指定字段

        :param id: 数据ID
        :param data: 更新的数据内容
        :param refresh: 强制刷新索引
        """
        kwargs = {}
        if refresh is not None:
            kwargs['refresh'] = refresh

        self.es.update(
            self.client.get_index_of_topic(self.topic),
            id=id,
            body={
                'doc': data,
            },
            **kwargs,
        )

    def search_by_dsl(self, query_body: dict) -> 'SearchResponse':
        resp = self.es.search(
            index=self.client.get_index_of_topic(self.topic),
            body=query_body,
        )
        return SearchResponse(resp)

    def count_by_dsl(self, query_body: dict) -> 'SearchResponse':
        resp = self.es.count(
            index=self.client.get_index_of_topic(self.topic),
            body=query_body,
        )
        return SearchResponse(resp)

    def update_by_dsl(self, query_body: dict, refresh: bool = None) -> 'BaseResponse':
        kwargs = {}
        if refresh is not None:
            kwargs['refresh'] = refresh

        resp = self.es.update_by_query(
            self.client.get_index_of_topic(self.topic),
            body=query_body,
            **kwargs,
        )
        return BaseResponse(resp)

    def bulk(self, actions: List[dict], refresh: bool = None):
        kwargs = {}
        if refresh is not None:
            kwargs['refresh'] = refresh

        success, errors = helpers.bulk(
            self.es,
            index=self.client.get_index_of_topic(self.topic),
            actions=actions,
            **kwargs,
        )
        return BulkResponse(dict(success=success, errors=errors))

    def query(self, name: str = None) -> 'QueryBuilder':
        """
        基于筛选表达式检索数据，影响评分

        :param name: 筛选的字段名称
        """

        q = IndexedQuery(self.topic, self.client)
        return QueryBuilder(name, q, q._queries)

    def filter(self, name: str = None) -> 'QueryBuilder':
        """
        基于筛选表达式检索数据，不影响评分

        :param name: 筛选的字段名称
        """

        q = IndexedQuery(self.topic, self.client)
        return QueryBuilder(name, q, q._filters)


class IndexedQuery(EsIndex):
    def __init__(self, index: str, client: EsBasedSearch):
        super().__init__(index, client)

        self._queries: List[NamedClause] = []
        self._filters: List[NamedClause] = []
        self._page = None
        self._size = None
        self._sort_by = []

    def query(self, name: str = None) -> 'QueryBuilder':
        return QueryBuilder(name, self, self._queries)

    def filter(self, name: str = None) -> 'QueryBuilder':
        return QueryBuilder(name, self, self._queries)

    def paginate(self, page: int = 0, size: int = 10) -> 'IndexedQuery':
        """
        分页查询

        :param page: 分页的页码，从0开始
        :param size: 分页的大小
        """

        self._page = page
        self._size = size
        return self

    def sort_by(self, name: str = None, order: str = None) -> 'IndexedQuery':
        if order is None:
            order = 'ASC'
        order = order.upper()
        self._sort_by.append({
            name: order
        })
        return self

    def hits(self) -> List[dict]:
        """
        返回命中查询的所有数据，默认进行了分页。
        """

        body = self._body_for_search()
        return self.search_by_dsl(body).hits()

    def raw_hits(self) -> List[dict]:
        body = self._body_for_search()
        return self.search_by_dsl(body).raw_hits()

    def count(self) -> int:
        """
        返回命中查询的数据量
        """

        body = {
            'query': self._build_query(),
        }
        return self.count_by_dsl(body).count()

    def _build_query(self) -> dict:
        query = BoolQuery()
        for q in self._queries:
            q.build(query.must, query.must_not)
        for f in self._filters:
            f.build(query.filter, query.must_not)
        return {
            'bool': query.dict()
        }

    def _body_for_search(self) -> dict:
        body = {
            'query': self._build_query(),
        }
        if self._size is not None:
            body['size'] = self._size
            if self._page is not None:
                body['from'] = self._page * self._size
        if self._sort_by:
            body['sort'] = self._sort_by
        return body


class BoolQuery(BaseModel):
    must: List[Dict] = []
    filter: List[Dict] = []
    must_not: List[Dict] = []
    should: List[Dict] = []


class ClauseType:
    IS = object()
    IS_NOT = object()
    IS_ONE_OF = object()
    IS_BETWEEN = object()
    IS_NOT_BETWEEN = object()
    EXISTS = object()
    DOES_NOT_EXIST = object()
    MATCH = object()
    MUST = object()


class NamedClause(BaseModel):
    type: Any
    name: Optional[str]
    value: Any = None

    def build(self, must: list, must_not: list):
        if self.type == ClauseType.IS:
            assert isinstance(self.value, Sequence)
            if self.name is None:
                must.append({
                    'simple_query_string': {
                        'query': ' '.join([f'"{i}"' for i in self.value]),
                        'default_operator': 'and',
                    }
                })
            else:
                for i in self.value:
                    must.append(self._match_phrase(self.name, i))
        elif self.type == ClauseType.IS_NOT:
            assert isinstance(self.value, Sequence)
            if self.name is None:
                must_not.append({
                    'simple_query_string': {
                        'query': ' '.join([f'"{i}"' for i in self.value]),
                    }
                })
            else:
                for i in self.value:
                    must_not.append(self._match_phrase(self.name, i))
        elif self.type == ClauseType.IS_ONE_OF:
            assert isinstance(self.value, Sequence)
            b = BoolQuery()
            if self.name is None:
                for v in self.value:
                    b.should.append({
                        'simple_query_string': {
                            'query': f'"{v}"',
                        }
                    })
            else:
                for v in self.value:
                    b.should.append(self._match_phrase(self.name, v))
            must.append({
                'bool': b.dict(),
            })
        elif self.type in (ClauseType.IS_BETWEEN, ClauseType.IS_NOT_BETWEEN):
            assert isinstance(self.value, tuple)
            if self.name is not None:
                q = {}
                if self.value[0] is not None:
                    q['gte'] = self.value[0]
                if self.value[1] is not None:
                    q['lte'] = self.value[1]
                if q:
                    _q = {
                        'range': {
                            self.name: q
                        }
                    }
                    if self.type == ClauseType.IS_BETWEEN:
                        must.append(_q)
                    else:
                        must_not.append(_q)
        elif self.type == ClauseType.EXISTS:
            if self.name is not None:
                must.append(self._exists(self.name))
        elif self.type == ClauseType.DOES_NOT_EXIST:
            if self.name is not None:
                must_not.append(self._exists(self.name))
        elif self.type == ClauseType.MATCH:
            assert isinstance(self.value, Sequence)
            if self.name is None:
                must.append({
                    'simple_query_string': {
                        'query': ' '.join([i for i in self.value]),
                    }
                })
            else:
                for i in self.value:
                    must.append(self._match(self.name, i))
        elif self.type == ClauseType.MUST:
            assert isinstance(self.value, dict)
            must.append(self.value)

    @staticmethod
    def _match_phrase(name: str, value: Any) -> dict:
        if ',' in name:
            s = name.split(',')
            return {
                "multi_match": {
                    "type": "phrase",
                    "query": value,
                    "fields": [i.strip() for i in s]
                }
            }
        return {
            'match_phrase': {
                name: value
            }
        }

    @staticmethod
    def _match(name: str, value: Any) -> dict:
        if ',' in name:
            s = name.split(',')
            return {
                "multi_match": {
                    "query": value,
                    "fields": [i.strip() for i in s]
                }
            }
        return {
            'match': {
                name: value,
            }
        }

    @staticmethod
    def _exists(name: str) -> dict:
        return {
            'exists': {
                'field': name
            }
        }


class QueryBuilder:
    def __init__(self, name: Optional[str], query: IndexedQuery, query_set: list):
        self._name = name
        self._query = query
        self._query_set = query_set

    def is_(self, *values: Any) -> IndexedQuery:
        """
        是某个值

        :param values: 具体值的列表
        """

        self._query_set.append(
            NamedClause(
                type=ClauseType.IS,
                name=self._name,
                value=values,
            )
        )
        return self._query

    def is_not(self, *values: Any) -> IndexedQuery:
        """
        不是某个值

        :param values: 具体值的列表
        """

        self._query_set.append(
            NamedClause(
                type=ClauseType.IS_NOT,
                name=self._name,
                value=values
            )
        )
        return self._query

    def is_one_of(self, *values: Any) -> IndexedQuery:
        """
        是其中某个值

        :param values: 具体值的列表
        """

        self._query_set.append(
            NamedClause(
                type=ClauseType.IS_ONE_OF,
                name=self._name,
                value=values,
            )
        )
        return self._query

    def is_between(self, lower_bound: Any = None, upper_bound: Any = None) -> IndexedQuery:
        """
        在区间范围内

        :param lower_bound: 最低值
        :param upper_bound: 最高值
        """

        self._query_set.append(
            NamedClause(
                type=ClauseType.IS_BETWEEN,
                name=self._name,
                value=(lower_bound, upper_bound),
            )
        )
        return self._query

    def is_not_between(self, lower_bound: Any = None, upper_bound: Any = None) -> IndexedQuery:
        """
        不在区间范围内

        :param lower_bound: 最低值
        :param upper_bound: 最高值
        """

        self._query_set.append(
            NamedClause(
                type=ClauseType.IS_NOT_BETWEEN,
                name=self._name,
                value=(lower_bound, upper_bound),
            )
        )
        return self._query

    def exists(self) -> IndexedQuery:
        """
        字段存在
        """

        self._query_set.append(
            NamedClause(
                type=ClauseType.EXISTS,
                name=self._name,
            )
        )
        return self._query

    def does_not_exists(self) -> IndexedQuery:
        """
        字段不存在
        """

        self._query_set.append(
            NamedClause(
                type=ClauseType.DOES_NOT_EXIST,
                name=self._name,
            )
        )
        return self._query

    def match(self, *values: Any) -> IndexedQuery:
        """
        模糊匹配

        :param values: 具体值的列表
        """

        self._query_set.append(
            NamedClause(
                type=ClauseType.MATCH,
                name=self._name,
                value=values,
            )
        )
        return self._query

    def must(self, query: dict) -> IndexedQuery:
        """
        must表达式

        :param query: 表达式
        """

        self._query_set.append(
            NamedClause(
                type=ClauseType.MUST,
                name=self._name,
                value=query,
            )
        )
        return self._query



import io
from typing import Any

import oss2
from minio import Minio, S3Error

from kikyo import Kikyo
from kikyo.oss import OSS, Bucket


class MinioBasedOSS(OSS):
    def __init__(self, client: Kikyo):
        settings = client.settings.deep('minio')

        if not settings:
            return

        self.client = client
        self.bucket_prefix = settings.get('bucket_prefix', default='')
        self.minio = Minio(
            settings['endpoint'],
            access_key=settings['access_key'],
            secret_key=settings['secret_key'],
            secure=settings['secure'],
        )
        self.minio_address = f"{'https' if settings['secure'] else 'http'}://{settings['endpoint']}"

        client.add_component('minio_oss', self)

    def bucket(self, name: str) -> Bucket:
        if name.startswith(self.bucket_prefix):
            bucket_name = name
        else:
            bucket_name = f'{self.bucket_prefix}{name}'

        return MinioBucket(bucket_name, self)


class MinioBucket(Bucket):
    def __init__(self, name: str, oss: MinioBasedOSS):
        self._name = name
        self.minio = oss.minio
        self.minio_address = oss.minio_address

    def get_object_link(self, key: str) -> str:
        return f'{self.minio_address}/{self._name}/{key}'

    def put_object(self, key: str, data: Any):
        _data = None
        _length = None
        if isinstance(data, bytes):
            _length = len(data)
            _data = io.BytesIO(data)
        else:
            raise RuntimeError(f'Unsupported data type: {type(data)}')
        self.minio.put_object(self._name, key, _data, _length)

    def get_object(self, key: str) -> Any:
        resp = self.minio.get_object(self._name, key)
        return resp.data

    def object_exists(self, key: str) -> bool:
        try:
            self.minio.get_object_tags(self._name, key)
        except S3Error as e:
            if e.code == 'NoSuchKey':
                return False
            raise
        return True

    def remove_object(self, key: str):
        self.minio.remove_object(self._name, key)


class AliyunOSS(OSS):
    def __init__(self, client: Kikyo):
        settings = client.settings.deep('oss2')

        if not settings:
            return

        self.client = client
        self.bucket_prefix = settings.get('bucket_prefix', default='')

        self.auth = oss2.Auth(settings['access_key'], settings['access_key_secret'])
        self.endpoint = settings['endpoint']
        self.secure = settings.get('secure', False)

        client.add_component('aliyun_oss', self)

    def bucket(self, name: str) -> Bucket:
        if name.startswith(self.bucket_prefix):
            bucket_name = name
        else:
            bucket_name = f'{self.bucket_prefix}{name}'

        return AliyunBucket(bucket_name, self)


class AliyunBucket(Bucket):
    def __init__(self, name: str, oss: AliyunOSS):
        self.name = name
        self.oss = oss
        self.bucket = oss2.Bucket(
            self.oss.auth,
            self.oss.endpoint,
            self.name,
            connect_timeout=20,
        )

    def get_object_link(self, key: str) -> str:
        return f'{"https" if self.oss.secure else "http"}://{self.name}.{self.oss.endpoint}/{key}'

    def put_object(self, key: str, data: Any):
        self.bucket.put_object(key, data)

    def get_object(self, key: str) -> Any:
        resp = self.bucket.get_object(key)
        return resp.read()

    def object_exists(self, key: str) -> bool:
        return self.bucket.object_exists(key)

    def remove_object(self, key: str):
        self.bucket.delete_object(key)



import base64
import io

import requests
import yaml

from kikyo.client import Kikyo
from kikyo.settings import Settings


def configure_by_consul(config_url: str, **kwargs) -> Kikyo:
    """从Consul拉取YAML格式的配置文件

    :param config_url: 获取配置项的URL地址
    """
    resp = requests.get(config_url)
    resp.raise_for_status()

    settings = Settings()
    for data in resp.json():
        v = data['Value']
        if not v:
            continue
        s = base64.b64decode(v)
        conf: dict = yaml.safe_load(io.BytesIO(s))
        if 'kikyo' in conf:
            settings.merge(conf['kikyo'])
            break

    settings.merge(kwargs)
    return Kikyo(settings)



import copy


class Settings(dict):
    def __init__(self, __values=None, **kwargs):
        super().__init__()
        self.update(__values, **kwargs)

    def __getitem__(self, name):
        return self.get(name)

    def deep(self, prefix) -> 'Settings':
        s = prefix.split('.')
        obj = self
        for i in s:
            if isinstance(obj, dict):
                if i in obj:
                    obj = obj[i]
                else:
                    obj = None
            else:
                obj = None
            if obj is None:
                break
        if obj is None:
            return Settings()
        return obj

    def get(self, name, default=None):
        return super().get(name, default)

    def getbool(self, name, default=None):
        v = self.get(name, default)
        return getbool(v)

    def getint(self, name, default=None):
        v = self.get(name, default)
        return getint(v)

    def getfloat(self, name, default=None):
        v = self.get(name, default)
        return getfloat(v)

    def getlist(self, name, default=None):
        v = self.get(name, default)
        return getlist(v)

    def set(self, name, value):
        self[name] = value

    def setdefault(self, k, default=None):
        if k not in self:
            self[k] = default

    def update(self, __values=None, **kwargs):
        if __values is not None:
            for name in __values:
                if isinstance(__values[name], dict):
                    self[name] = Settings(__values[name])
                else:
                    self[name] = __values[name]
        for k, v in kwargs.items():
            self[k] = Settings(v) if isinstance(v, dict) else v

    def merge(self, __values=None):
        if __values is not None:
            for name in __values:
                if isinstance(__values[name], dict):
                    v = Settings(__values[name])
                else:
                    v = __values[name]

                if name not in self:
                    self[name] = v
                else:
                    if isinstance(self[name], Settings) and isinstance(v, dict):
                        self[name].merge(v)
                    else:
                        self[name] = v

    def delete(self, name):
        del self[name]

    def copy(self):
        return copy.deepcopy(self)


def getbool(v):
    try:
        return bool(int(v))
    except (ValueError, TypeError):
        if v in ("True", "true"):
            return True
        if v in ("False", "false"):
            return False
    return None


def getint(v):
    try:
        return int(v)
    except (ValueError, TypeError):
        pass
    return None


def getfloat(v):
    try:
        return float(v)
    except (ValueError, TypeError):
        pass
    return None


def getlist(v):
    if v is None:
        return None
    if isinstance(v, str):
        v = v.split(",")
    elif not hasattr(v, "__iter__"):
        v = [v]
    return list(v)
